# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from pathlib import Path
from typing import Callable, IO, NamedTuple, overload

from .kgmvprdataciphers import KGMCryptoLegacy
from ..exceptions import CrypterCreatingError
from ..keyutils import make_salt
from ..miscutils import proberfuncfactory
from ..prototypes import EncryptedBytesIO
from ..typedefs import BytesLike, FilePath, KeyStreamBasedStreamCipherProto, StreamCipherProto
from ..typeutils import isfilepath, tobytes, verify_fileobj

warnings.filterwarnings(action='default', category=DeprecationWarning, module=__name__)

__all__ = ['KGMorVPR', 'probe_kgmvpr', 'KGMorVPRFileInfo']


class KGMorVPRFileInfo(NamedTuple):
    cipher_data_offset: int
    """加密数据在文件中开始的位置。"""
    cipher_data_len: int
    """加密数据在文件中的长度。"""
    encryption_version: int
    """加密方法的版本。目前仅支持 3。"""
    core_key_slot: int
    """加解密数据所需的密钥槽序号。"""
    core_key_test_data: bytes
    """用于验证文件内主密钥合法性的数据。"""
    master_key: bytes | None
    """主密钥。需要配合 ``core_key_slot`` 对应的密钥使用。"""
    is_vpr: bool
    """如果文件使用了 VPR 加密，则为 ``True``。"""
    opener: Callable[[tuple[FilePath | IO[bytes], KGMorVPRFileInfo] | FilePath | IO[bytes], ...], KGMorVPR]
    """打开文件的方式，为一个可调对象，其会返回一个加密文件对象。"""
    opener_kwargs_required: tuple[str, ...]
    """通过 ``opener`` 打开文件时，所必需的关键字参数的名称。"""
    opener_kwargs_optional: tuple[str, ...]
    """通过 ``opener`` 打开文件时，可选的关键字参数的名称。

    此属性仅储存可能会影响 ``opener`` 行为的可选关键字参数；
    对 ``opener`` 行为没有影响的可选关键字参数不会出现在此属性中。
    """


@overload
def probe_kgmvpr(filething: FilePath | IO[bytes], /) -> tuple[Path | IO[bytes], KGMorVPRFileInfo | None]:
    pass


@proberfuncfactory
def probe_kgmvpr(filething, /):
    """探测源文件 ``filething`` 是否为一个 KGM 或 VPR 文件。

    返回一个 2 个元素长度的元组：第一个元素为 ``filething``；如果
    ``filething`` 是 KGM 或 VPR 文件，那么第二个元素为一个 ``KGMorVPRFileInfo`` 对象；否则为 ``None``。

    如果 ``filething`` 是 VPR 文件，那么 ``KGMorVPRFileInfo``
    对象的 ``is_vpr`` 属性为 ``True``；否则为 ``False``。

    本方法的返回值可以用于 ``KGMorVPR.open()`` 的第一个位置参数。

    Args:
        filething: 源文件的路径或文件对象
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 KGM 或 VPR 文件，那么第二个元素为一个 KGMorVPRFileInfo 对象；否则为 None。
    """
    opener_kwargs_required = ['table1', 'table2', 'tablev2']

    total_size = filething.seek(0, 2)
    if total_size < 60:
        return
    filething.seek(0, 0)

    header = filething.read(16)
    if header == b'\x05\x28\xbc\x96\xe9\xe4\x5a\x43\x91\xaa\xbd\xd0\x7a\xf5\x36\x31':
        is_vpr = True
        opener_kwargs_required.append('vpr_key')
    elif header == b'\x7c\xd5\x32\xeb\x86\x02\x7f\x4b\xa8\xaf\xa6\x8e\x0f\xff\x99\x14':
        is_vpr = False
    else:
        return

    cipher_data_offset = int.from_bytes(filething.read(4), 'little')
    encryption_version = int.from_bytes(filething.read(4), 'little')
    core_key_slot = int.from_bytes(filething.read(4), 'little')
    core_key_test_data = filething.read(16)
    master_key = filething.read(16)

    return KGMorVPRFileInfo(
        cipher_data_offset=cipher_data_offset,
        cipher_data_len=total_size - cipher_data_offset,
        encryption_version=encryption_version,
        core_key_slot=core_key_slot,
        core_key_test_data=core_key_test_data,
        master_key=master_key,
        is_vpr=is_vpr,
        opener=KGMorVPR.open,
        opener_kwargs_required=tuple(opener_kwargs_required),
        opener_kwargs_optional=()
    )


class KGMorVPR(EncryptedBytesIO):
    """基于 BytesIO 的 KGM/VPR 透明加密二进制流。

    所有读写相关方法都会经过透明加密层处理：
    读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

    调用读写相关方法时，附加参数 ``nocryptlayer=True``
    可绕过透明加密层，访问缓冲区内的原始加密数据。

    如果你要新建一个 KGMorVPR 对象，不要直接调用 ``__init__()``，而是使用构造器方法
    ``KGMorVPR.new()`` 和 ``KGMorVPR.open()`` 新建或打开已有 KGM/VPR 文件，
    使用已有 KGMorVPR 对象的 ``save()`` 方法将其保存到文件。
    """

    @property
    def acceptable_ciphers(self):
        return [KGMCryptoLegacy]

    def __init__(self,
                 cipher: StreamCipherProto | KeyStreamBasedStreamCipherProto, /,
                 initial_bytes: BytesLike = b''
                 ) -> None:
        """基于 BytesIO 的 KGM/VPR 透明加密二进制流。

        所有读写相关方法都会经过透明加密层处理：
        读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

        调用读写相关方法时，附加参数 ``nocryptlayer=True``
        可绕过透明加密层，访问缓冲区内的原始加密数据。

        如果你要新建一个 KGMorVPR 对象，不要直接调用 ``__init__()``，而是使用构造器方法
        ``KGMorVPR.new()`` 和 ``KGMorVPR.open()`` 新建或打开已有 KGM/VPR 文件，
        使用已有 KGMorVPR 对象的 ``save()`` 方法将其保存到文件。

        Args:
            cipher: 要使用的 cipher，必须是一个 libtakiyasha.kgmvpr.kgmvprdataciphers.KGMCryptoLegacy 对象
            initial_bytes: 内置缓冲区的初始数据
        """
        super().__init__(cipher, initial_bytes)

        self._source_file_header_data: bytes | None = None

    @classmethod
    def from_file(cls,
                  kgm_vpr_filething: FilePath | IO[bytes], /,
                  table1: BytesLike,
                  table2: BytesLike,
                  tablev2: BytesLike,
                  vpr_key: BytesLike = None
                  ):
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``KGMorVPR.open()`` 代替。）

        打开一个 KGMorVPR 文件或文件对象 ``kgm_vpr_filething``。

        第一个位置参数 ``kgm_vpr_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``kgm_vpr_filething``
        也可以是一个文件对象，但必须可读、可跳转（``kgm_vpr_filething.seekable() == True``）。

        参数 ``table1``、``table2``、``tablev2`` 都是必选参数，
        因为它们会参与到内置透明加密层的创建过程中，并且在加密/解密过程中发挥关键作用。
        这三个参数的都必须是类字节对象，且转换为 ``bytes`` 后，长度为 272 字节。

        本方法会寻找文件内嵌主密钥的位置和加密方式，进而判断所用加密算法的类型。

        如果探测到 ``VPR`` 文件，那么参数 ``vpr_key`` 是必选的：必须是类字节对象，且转换为 ``bytes``
        后的长度为 17 字节。
        """
        warnings.warn(
            DeprecationWarning(
                f'{cls.__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {cls.__name__}.open() instead.'
            )
        )
        return cls.open(kgm_vpr_filething,
                        table1=table1,
                        table2=table2,
                        tablev2=tablev2,
                        vpr_key=vpr_key
                        )

    @classmethod
    def open(cls,
             filething_or_info: tuple[Path | IO[bytes]] | FilePath | IO[bytes], /,
             table1: BytesLike,
             table2: BytesLike,
             tablev2: BytesLike,
             vpr_key: BytesLike = None
             ):
        """打开一个 KGMorVPR 文件，并返回一个 ``KGMorVPR`` 对象。

        第一个位置参数 ``filething_or_info`` 需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        ``filething_or_info`` 也可以接受 ``probe_kgmvpr()`` 函数的返回值：
        一个包含两个元素的元组，第一个元素是源文件的路径或文件对象，第二个元素是源文件的信息。

        第二、三、四个参数 ``table1``、``table2`` 和 ``tablev2``
        是必需的，都必须是 272 字节长度的字节串。

        如果探测到 VPR 文件，那么第五个参数 ``vpr_key``
        是必需的。如果提供，则必须是 17 字节长度的字节串。其他情况下，此参数会被忽略。

        Args:
            filething_or_info: 源文件的路径或文件对象，或者 probe_kgmvpr() 的返回值
            table1: 解码表 1
            table2: 解码表 2
            tablev2: 解码表 3
            vpr_key: 针对 VPR 文件额外所需的密钥
        """
        table1 = tobytes(table1)
        table2 = tobytes(table2)
        tablev2 = tobytes(tablev2)
        if vpr_key is not None:
            vpr_key = tobytes(vpr_key)

        def operation(fd: IO[bytes]) -> cls:
            if fileinfo.encryption_version != 3:
                raise CrypterCreatingError(
                    f'unsupported KGM encryption version {fileinfo.encryption_version} '
                    f'(only version 3 is supported)'
                )
            if fileinfo.is_vpr and vpr_key is None:
                raise TypeError(
                    "argument 'vpr_key' is required for encrypt and decrypt VPR file"
                )
            cipher = KGMCryptoLegacy(table1,
                                     table2,
                                     tablev2,
                                     fileinfo.core_key_test_data + b'\x00',
                                     vpr_key
                                     )

            fd.seek(fileinfo.cipher_data_offset, 0)

            inst = cls(cipher, fd.read(fileinfo.cipher_data_len))
            fd.seek(0, 0)
            inst._source_file_header_data = fd.read(fileinfo.cipher_data_offset)
            return inst

        if isinstance(filething_or_info, tuple):
            filething_or_info: tuple[Path | IO[bytes], KGMorVPRFileInfo | None]
            if len(filething_or_info) != 2:
                raise TypeError(
                    "first argument 'filething_or_info' must be a file path, a file object, "
                    "or a tuple of probe_kgmvpr() returns"
                )
            filething, fileinfo = filething_or_info
        else:
            filething, fileinfo = probe_kgmvpr(filething_or_info)

        if fileinfo is None:
            raise CrypterCreatingError(
                f"{repr(filething)} is not a KGM or VPR file"
            )
        elif not isinstance(fileinfo, KGMorVPRFileInfo):
            raise TypeError(
                f"second element of the tuple must be KGMorVPRFileInfo or None, not {type(fileinfo).__name__}"
            )

        if isfilepath(filething):
            with open(filething, mode='rb') as fileobj:
                instance = operation(fileobj)
                instance._name = Path(filething)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_readable=True,
                                     verify_seekable=True
                                     )
            fileobj_sourcefile = getattr(fileobj, 'name', None)
            instance = operation(fileobj)

            if fileobj_sourcefile is not None:
                instance._name = Path(fileobj_sourcefile)

        return instance

    def to_file(self, kgm_vpr_filething: FilePath | IO[bytes] = None) -> None:
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``KGMorVPR.save()`` 代替。）

        将当前 KGMorVPR 对象的内容保存到文件 ``kgm_vpr_filething``。

        第一个位置参数 ``kgm_vpr_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``kgm_vpr_filething``
        也可以是一个文件对象，但必须可写。

        本方法会首先尝试写入 ``kgm_vpr_filething`` 指向的文件。
        如果未提供 ``kgm_vpr_filething``，则会尝试写入 ``self.name``
        指向的文件。如果两者都为空或未提供，则会触发 ``CrypterSavingError``。

        目前无法生成 KGM/VPR 文件的标头数据，因此本方法不能用于保存通过 ``KGMorVPR.new()``
        创建的 ``KGMorVPR`` 对象。尝试这样做会触发 ``NotImplementedError``。
        """
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.save() instead.'
            )
        )

        return self.save(kgm_vpr_filething)

    def save(self,
             filething: FilePath | IO[bytes] = None
             ) -> None:
        """（实验性功能）将当前对象保存为一个新 KGM 或 VPR 文件。

        第一个参数 ``filething`` 是可选的，如果提供此参数，需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。
        如果未提供此参数，那么将会尝试使用当前对象的 ``source`` 属性；如果后者也不可用，则引发
        ``TypeError``。

        目前无法生成 KGM/VPR 文件的标头数据，因此本方法不能用于保存通过 ``KGMorVPR.new()``
        创建的 ``KGMorVPR`` 对象。尝试这样做会触发 ``NotImplementedError``。

        Args:
            filething: 目标文件的路径或文件对象
        """

        def operation(fd: IO[bytes]) -> None:
            if self._source_file_header_data is None:
                raise NotImplementedError(
                    f"cannot save current {type(self).__name__} object to file '{str(filething)}', "
                    f"generate KGM/VPR file header is not supported"
                )
            fd.seek(0, 0)
            fd.write(self._source_file_header_data)
            while blk := self.read(self.DEFAULT_BUFFER_SIZE, nocryptlayer=True):
                fd.write(blk)

        if filething is None:
            if self.source is None:
                raise TypeError(
                    "attribute 'self.source' and argument 'filething' are empty, "
                    "don't know which file to save to"
                )
            filething = self.source

        if isfilepath(filething):
            with open(filething, mode='wb') as fileobj:
                return operation(fileobj)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_seekable=True,
                                     verify_writable=True
                                     )
            return operation(fileobj)

    @classmethod
    def new(cls,
            table1: BytesLike,
            table2: BytesLike,
            tablev2: BytesLike,
            vpr_key: BytesLike = None
            ):
        """返回一个空 KGMorVPR 对象。

        第一、二、三个参数 ``table1``、``table2`` 和 ``tablev2``
        是必需的，都必须是 272 字节长度的字节串。

        如果提供了第五个参数 ``vpr_key``，那么将会使用针对 VPR 的加密方法；这种情况下
        ``vpr_key`` 必须是 17 字节长度的字节串。

        注意：通过本方法创建的 ``KGMorVPR`` 对象不可通过 ``save()``
        方法保存到文件。尝试这样做会触发 ``NotImplementedError``。
        """
        table1 = tobytes(table1)
        table2 = tobytes(table2)
        tablev2 = tobytes(tablev2)
        if vpr_key is not None:
            vpr_key = tobytes(vpr_key)

        core_key_test_data = make_salt(16) + b'\x00'

        cipher = KGMCryptoLegacy(table1, table2, tablev2, core_key_test_data, vpr_key)

        return cls(cipher)



# -*- coding: utf-8 -*-
from __future__ import annotations

from functools import lru_cache
from typing import Generator, Literal

from .qmcconsts import KEY256_MAPPING
from ..prototypes import KeyStreamBasedStreamCipherSkel
from ..typedefs import BytesLike, IntegerLike
from ..typeutils import CachedClassInstanceProperty, tobytes, toint


class Mask128(KeyStreamBasedStreamCipherSkel):
    def __init__(self, mask128: BytesLike, /):
        self._mask128 = tobytes(mask128)
        if len(self._mask128) != 128:
            raise ValueError(f"invalid mask length: should be 128, got {len(self._mask128)}")

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._mask128
        elif keyname == 'original':
            return getattr(self, '_original_qmcv2_key256', None)

    @classmethod
    def cls_keystream(cls,
                      mask128: BytesLike,
                      nbytes: IntegerLike,
                      offset: IntegerLike, /
                      ) -> Generator[int, None, None]:
        mask = tobytes(mask128)
        if len(mask) != 128:
            raise ValueError(f"invalid mask length: should be 128, got {len(mask)}")
        nbytes = toint(nbytes)
        offset = toint(offset)
        if offset < 0:
            raise ValueError("third argument 'offset' must be a non-negative integer")
        if nbytes < 0:
            raise ValueError("second argument 'nbytes' must be a non-negative integer")

        firstblk_data = mask * 256  # 前 32768 字节
        secondblk_data = firstblk_data[1:-1]  # 第 32769 至 65535 字节
        startblk_data = firstblk_data + secondblk_data  # 初始块：前 65535 字节
        startblk_len = len(startblk_data)
        commonblk_data = firstblk_data[:-1]  # 普通块：第 65536 字节往后每一个 32767 大小的块
        commonblk_len = len(commonblk_data)

        if 0 <= offset < startblk_len:
            max_target_in_startblk_len = startblk_len - offset
            target_in_commonblk_len = nbytes - max_target_in_startblk_len
            target_in_startblk_len = min(nbytes, max_target_in_startblk_len)
            yield from startblk_data[offset:offset + target_in_startblk_len]
            if target_in_commonblk_len <= 0:
                return
            else:
                offset = 0
        else:
            offset -= startblk_len
            target_in_commonblk_len = nbytes

        target_offset_in_commonblk = offset % commonblk_len
        if target_offset_in_commonblk == 0:
            target_before_commonblk_area_len = 0
        else:
            target_before_commonblk_area_len = commonblk_len - target_offset_in_commonblk
        yield from commonblk_data[target_offset_in_commonblk:target_offset_in_commonblk + target_before_commonblk_area_len]
        target_in_commonblk_len -= target_before_commonblk_area_len

        target_overrided_whole_commonblk_count = target_in_commonblk_len // commonblk_len
        target_after_commonblk_area_len = target_in_commonblk_len % commonblk_len

        for _ in range(target_overrided_whole_commonblk_count):
            yield from commonblk_data
        yield from commonblk_data[:target_after_commonblk_area_len]

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        yield from self.cls_keystream(self._mask128, nbytes, offset)

    @classmethod
    def from_qmcv1_mask44(cls, mask44: BytesLike) -> Mask128:
        mask44: bytes = tobytes(mask44)
        if len(mask44) != 44:
            raise ValueError(f'invalid mask length (should be 44, got {len(mask44)})')

        mask128 = bytearray(128)
        idx44 = 0
        for it256 in KEY256_MAPPING:
            if it256:
                for idx128 in it256:
                    mask128[idx128] = mask44[idx44]
                idx44 += 1

        return cls(mask128)

    @classmethod
    def from_qmcv1_mask256(cls, mask256: BytesLike) -> Mask128:
        mask256: bytes = tobytes(mask256)
        if len(mask256) != 256:
            raise ValueError(f'invalid mask length (should be 256, got {len(mask256)})')

        mask128 = bytearray(128)
        for idx128 in range(128):
            if idx128 > 0x7fff:
                idx128 %= 0x7fff
            idx = (idx128 ** 2 + 27) & 0xff
            mask128[idx128] = mask256[idx]

        return cls(mask128)

    @classmethod
    def from_qmcv2_key256(cls, key256: BytesLike) -> Mask128:
        key256: bytes = tobytes(key256)
        if len(key256) != 256:
            raise ValueError(f'invalid key length (should be 256, got {len(key256)})')

        mask128 = bytearray(128)
        for idx128 in range(128):
            if idx128 > 0x7fff:
                idx128 %= 0x7fff
            idx = (idx128 ** 2 + 71214) % 256

            value = key256[idx]
            rotate = ((idx & 7) + 4) % 8

            mask128[idx128] = ((value << rotate) % 256) | ((value >> rotate) % 256)

        instance = cls(mask128)
        instance._original_qmcv2_key256 = key256

        return instance


class HardenedRC4(KeyStreamBasedStreamCipherSkel):
    def __init__(self, key: BytesLike, /):
        self._key = tobytes(key)

        key_len = len(self._key)
        if key_len == 0:
            raise ValueError("first argument 'key' cannot be an empty bytestring")
        if b'\x00' in self._key:
            raise ValueError("first argument 'key' cannot contain null bytes")

        self._box = bytearray(i % 256 for i in range(key_len))
        j = 0
        for i in range(key_len):
            j = (j + self._box[i] + self._key[i]) % key_len
            self._box[i], self._box[j] = self._box[j], self._box[i]

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._key

    @property
    @lru_cache
    def hash_base(self) -> int:
        base = 1
        key = self._key

        for i in range(len(key)):
            v: int = key[i]
            if v == 0:
                continue
            next_hash: int = (base * v) & 0xffffffff
            if next_hash == 0 or next_hash <= base:
                break
            base = next_hash
        return base

    @CachedClassInstanceProperty
    def first_segment_size(self) -> int:
        return 128

    @CachedClassInstanceProperty
    def common_segment_size(self) -> int:
        return 5120

    @lru_cache(maxsize=65536)
    def _get_segment_skip(self, value: int) -> int:
        key = self._key
        key_len = len(self._key)

        seed = key[value % key_len]
        idx = int(self.hash_base / ((value + 1) * seed) * 100)

        return idx % key_len

    def _yield_first_segment_keystream(self,
                                       blksize: int,
                                       offset: int
                                       ) -> Generator[int, None, None]:
        key = self._key
        for i in range(offset, offset + blksize):
            yield key[self._get_segment_skip(i)]

    def _yield_common_segment_keystream(self,
                                        blksize: int,
                                        offset: int
                                        ) -> Generator[int, None, None]:
        key_len = len(self._key)
        box = self._box.copy()
        j, k = 0, 0

        skip_len = offset % self.common_segment_size + self._get_segment_skip(
            offset // self.common_segment_size
        )
        for i in range(-skip_len, blksize):
            j = (j + 1) % key_len
            k = (box[j] + k) % key_len
            box[j], box[k] = box[k], box[j]
            if i >= 0:
                yield box[(box[j] + box[k]) % key_len]

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        common_segment_size: int = self.common_segment_size

        pending = toint(nbytes)
        done = 0
        offset = toint(offset)
        if offset < 0:
            raise ValueError("third argument 'offset' must be a non-negative integer")
        if pending < 0:
            raise ValueError("second argument 'nbytes' must be a non-negative integer")

        if 0 <= offset < self.first_segment_size:
            blksize = min(pending, self.first_segment_size - offset)
            yield from self._yield_first_segment_keystream(blksize, offset)
            pending -= blksize
            done += blksize
            offset += blksize
            if pending <= 0:
                return

        if offset % common_segment_size != 0:
            blksize = min(pending, common_segment_size - (offset % common_segment_size))
            yield from self._yield_common_segment_keystream(blksize, offset)
            pending -= blksize
            done += blksize
            offset += blksize
            if pending <= 0:
                return

        while pending > common_segment_size:
            yield from self._yield_common_segment_keystream(common_segment_size, offset)
            pending -= common_segment_size
            done += common_segment_size
            offset += common_segment_size

        if pending > 0:
            yield from self._yield_common_segment_keystream(nbytes - done, offset)



# -*- coding: utf-8 -*-
from __future__ import annotations

from functools import partial
from secrets import randbelow as srandbelow

try:
    import io
except ImportError:
    import _pyio as io
from typing import Generator, Literal

from pyaes import AESModeOfOperationECB
from pyaes.util import append_PKCS7_padding, strip_PKCS7_padding

from .prototypes import KeyStreamBasedStreamCipherSkel, CipherSkel
from .exceptions import CipherDecryptingError
from .typedefs import IntegerLike, BytesLike
from .miscutils import bytestrxor
from .typeutils import CachedClassInstanceProperty, tobytes, toint

__all__ = [
    'StreamedAESWithModeECB',
    'TEAWithModeECB',
    'TarsCppTCTEAWithModeCBC',
    'ARC4'
]

rand = partial(srandbelow, 256)


class StreamedAESWithModeECB(CipherSkel):
    @CachedClassInstanceProperty
    def blocksize(self) -> int:
        return 16

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._key

    def __init__(self, key: BytesLike, /) -> None:
        self._key = tobytes(key)
        self._base_aes_cipher = AESModeOfOperationECB(self._key)

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        plaindata_padded = append_PKCS7_padding(tobytes(plaindata))
        iterable_by_blksize = iter(partial(io.BytesIO(plaindata_padded).read, self.blocksize), b'')
        return b''.join(self._base_aes_cipher.encrypt(blk) for blk in iterable_by_blksize)

    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        cipherdata = tobytes(cipherdata)
        iterable_by_blksize = iter(partial(io.BytesIO(cipherdata).read, self.blocksize), b'')
        plaindata_padded = b''.join(self._base_aes_cipher.decrypt(blk) for blk in iterable_by_blksize)
        return strip_PKCS7_padding(plaindata_padded)


class TEAWithModeECB(CipherSkel):
    @CachedClassInstanceProperty
    def blocksize(self) -> int:
        return 16

    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._key

    def __init__(self,
                 key: BytesLike,
                 /,
                 rounds: IntegerLike = 64,
                 magic_number: IntegerLike = 0x9e3779b9
                 ) -> None:
        self._key = tobytes(key)
        self._rounds = toint(rounds)
        self._delta = toint(magic_number)

        if len(self._key) != self.blocksize:
            raise ValueError(f"invalid key length: should be {self.blocksize}, not {len(self._key)}")
        if self._rounds % 2 != 0:
            raise ValueError(f'an even number of rounds is required (but got {self._rounds})')

    @classmethod
    def transvalues(cls, data: bytes, key: bytes) -> tuple[int, int, int, int, int, int]:
        v0 = int.from_bytes(data[:4], 'big')
        v1 = int.from_bytes(data[4:8], 'big')
        k0 = int.from_bytes(key[:4], 'big')
        k1 = int.from_bytes(key[4:8], 'big')
        k2 = int.from_bytes(key[8:12], 'big')
        k3 = int.from_bytes(key[12:], 'big')

        return v0, v1, k0, k1, k2, k3

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        master_key = self.getkey('master')
        v0, v1, k0, k1, k2, k3 = self.transvalues(tobytes(plaindata), master_key)

        delta = self._delta
        rounds = self._rounds
        ciphersum = 0

        for i in range(rounds // 2):
            ciphersum += delta
            ciphersum &= 0xffffffff
            v0 += ((v1 << 4) + k0) ^ (v1 + ciphersum) ^ ((v1 >> 5) + k1)
            v0 &= 0xffffffff
            v1 += ((v0 << 4) + k2) ^ (v0 + ciphersum) ^ ((v0 >> 5) + k3)
            v1 &= 0xffffffff

        return v0.to_bytes(4, 'big') + v1.to_bytes(4, 'big')

    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        master_key = self.getkey('master')
        v0, v1, k0, k1, k2, k3 = self.transvalues(tobytes(cipherdata), master_key)

        delta = self._delta
        rounds = self._rounds
        ciphersum = (delta * (rounds // 2)) & 0xffffffff

        for i in range(rounds // 2):
            v1 -= ((v0 << 4) + k2) ^ (v0 + ciphersum) ^ ((v0 >> 5) + k3)
            v1 &= 0xffffffff
            v0 -= ((v1 << 4) + k0) ^ (v1 + ciphersum) ^ ((v1 >> 5) + k1)
            v0 &= 0xffffffff
            ciphersum -= delta
            ciphersum &= 0xffffffff

        return v0.to_bytes(4, 'big') + v1.to_bytes(4, 'big')


class TarsCppTCTEAWithModeCBC(CipherSkel):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._lower_level_tea_cipher.getkey('master')

    @CachedClassInstanceProperty
    def blocksize(self) -> int:
        return 8

    @CachedClassInstanceProperty
    def master_key_size(self) -> int:
        return 16

    @CachedClassInstanceProperty
    def salt_len(self) -> int:
        return 2

    @CachedClassInstanceProperty
    def zero_len(self) -> int:
        return 7

    @property
    def lower_level_cipher(self) -> TEAWithModeECB:
        """使用的下层 Cipher。"""
        return self._lower_level_tea_cipher

    def __init__(self,
                 key: BytesLike,
                 /,
                 rounds: IntegerLike = 64,
                 magic_number: IntegerLike = 0x9e3779b9,
                 ) -> None:
        """TarsCpp 的 util/tc_tea 的完整实现。

        Args:
            key: 密钥，长度必须等于 16
            rounds: 加/解密的轮转次数，必须为偶数
            magic_number: 加/解密使用的魔数
        """
        key = tobytes(key)
        rounds = toint(rounds)
        magic_number = toint(magic_number)
        if len(key) != self.master_key_size:
            raise ValueError(f"invalid key length {len(key)}: "
                             f"should be {self.master_key_size}, not {len(key)}"
                             )
        if rounds % 2 != 0:
            raise ValueError(f"'rounds' must be an even integer, not {rounds}")

        self._lower_level_tea_cipher = TEAWithModeECB(key, rounds, magic_number)

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        # 根据 plaindata 长度计算 pad_len，最小长度必须为 8 的整数倍
        plaindata = tobytes(plaindata)

        pad_salt_body_zero_len = (len(plaindata) + self.salt_len + self.zero_len + 1)
        pad_len = pad_salt_body_zero_len % self.blocksize
        if pad_len != 0:
            # 模 8 余 0 需补 0，余 1 补 7，余 2 补 6，...，余 7 补 1
            pad_len = self.blocksize - pad_len

        src_buf = bytearray(self.blocksize)

        # 加密第一块数据（8 个字节）
        src_buf[0] = (rand() & 0xf8)  # 最低三位存 pad_len，清零
        src_buf[0] |= pad_len
        src_idx = 1  # src_idx 指向 src_buf 下一个位置

        # 填充
        while pad_len > 0:
            src_buf[src_idx] = rand()
            src_idx += 1
            pad_len -= 1

        # 到此处为止，src_idx 必须小于 8

        iv_plain = bytearray(self.blocksize)
        iv_crypt = iv_plain[:]  # 制造一个空初始向量

        # 获取加密结果预期长度，并据此创建一个空数组
        out_buf_len = self.get_encrypt_result_len(plaindata)
        out_buf = bytearray(out_buf_len)
        out_buf_pos = 0

        def crypt_block():  # CBC 加密操作流程
            nonlocal src_idx, out_buf_pos
            # 加密前异或前 8 个字节的密文（iv_crypt 指向的）
            src_buf[:] = bytestrxor(src_buf, iv_crypt)

            # 使用 TEA ECB 模式加密
            out_buf[out_buf_pos:out_buf_pos + self.blocksize] = self._lower_level_tea_cipher.encrypt(src_buf)

            # 加密后异或前8个字节的密文（iv_crypt 指向的）
            out_buf[out_buf_pos:out_buf_pos + self.blocksize] = bytestrxor(
                out_buf[out_buf_pos:out_buf_pos + self.blocksize], iv_plain
            )

            # 保存当前的 iv_plain
            iv_plain[:] = src_buf[:]

            # 更新 iv_crypt
            iv_crypt[:] = out_buf[out_buf_pos:out_buf_pos + self.blocksize]
            out_buf_pos += self.blocksize

        # 填充2个字节的 Salt
        i = 1
        while i <= self.salt_len:
            if src_idx < self.blocksize:
                src_buf[src_idx] = rand()
                src_idx += 1
                i += 1
            if src_idx == self.blocksize:
                crypt_block()
                src_idx = 0

        # src_idx 指向 src_buf 下一个位置

        plaindata_pos = 0
        while plaindata_pos < len(plaindata):
            if src_idx < self.blocksize:
                src_buf[src_idx] = plaindata[plaindata_pos]
                src_idx += 1
                plaindata_pos += 1
            if src_idx == self.blocksize:
                crypt_block()
                src_idx = 0

        # src_idx 指向 src_buf 下一个位置

        i = 1
        while i <= self.zero_len:
            if src_idx < 8:
                src_buf[src_idx] = 0
                src_idx += 1
                i += 1
            if src_idx == 8:
                crypt_block()
                src_idx = 0

        return bytes(out_buf)

    def get_encrypt_result_len(self, plaindata: bytes) -> int:
        # 根据 plaindata 长度计算 pad_len ，最小长度必须为8的整数倍
        pad_salt_body_zero_len = (len(plaindata) + self.salt_len + self.zero_len + 1)
        pad_len = pad_salt_body_zero_len % self.blocksize
        if pad_len != 0:
            # 模8余0需补0，余1补7，余2补6，...，余7补1
            pad_len = self.blocksize - pad_len

        # 返回的是加密结果预期的长度
        return pad_salt_body_zero_len + pad_len

    def decrypt(self,
                cipherdata: BytesLike, /,
                zero_check: bool = False
                ) -> bytes:
        cipherdata = tobytes(cipherdata)

        if len(cipherdata) < self.blocksize * 2:
            raise ValueError(
                f'invalid cipherdata length: should be greater than '
                f'{self.blocksize * 2}, not {len(cipherdata)}'
            )
        if len(cipherdata) % self.blocksize != 0:
            raise ValueError(
                f'invalid cipherdata length ({len(cipherdata)}): '
                f'not a integer multiple of {self.blocksize}'
            )

        dest_buf = bytearray(self._lower_level_tea_cipher.decrypt(cipherdata[:8]))
        pad_len = dest_buf[0] & 0x7
        out_buf_len = len(cipherdata) - pad_len - self.salt_len - self.zero_len - 1
        if out_buf_len < 0:
            raise CipherDecryptingError(
                'estimated plaindata length is less than 0'
            )
        out_buf = bytearray(out_buf_len)

        iv_previous = bytearray(8)
        iv_current = bytearray(cipherdata[:8])

        cipherdata_pos = 8
        dest_idx = 1 + pad_len

        def crypt_block() -> None:
            nonlocal cipherdata_pos
            iv_previous[:] = iv_current[:]
            iv_current[:] = cipherdata[cipherdata_pos:cipherdata_pos + 8]
            dest_buf[:] = self._lower_level_tea_cipher.decrypt(
                bytestrxor(
                    dest_buf[:8], iv_current[:8]
                )
            )
            cipherdata_pos += 8

        i = 1
        while i <= self.salt_len:
            if dest_idx < 8:
                dest_idx += 1
                i += 1
            elif dest_idx == 8:
                crypt_block()
                dest_idx = 0

        out_buf_pos = 0
        while out_buf_pos < out_buf_len:
            if dest_idx < 8:
                out_buf[out_buf_pos] = dest_buf[dest_idx] ^ iv_previous[dest_idx]
                dest_idx += 1
                out_buf_pos += 1
            elif dest_idx == 8:
                crypt_block()
                dest_idx = 0

        if zero_check:
            for i in range(1, self.zero_len):
                if dest_idx < 8:
                    if dest_buf[dest_idx] ^ iv_previous[dest_idx] != 0:
                        raise CipherDecryptingError('zero check failed')
                    dest_idx += 1
                elif dest_idx == 8:
                    crypt_block()
                    dest_idx = 0

        return bytes(out_buf)


class ARC4(KeyStreamBasedStreamCipherSkel):
    def __init__(self, key: BytesLike, /) -> None:
        """标准的 RC4 加密算法实现。

        Args:
            key: 密钥，长度不可大于 256，且不可为空
        """
        key = tobytes(key)
        self._key = key
        key_len = len(key)
        if key_len == 0:
            raise ValueError(f"first argument 'key' cannot be an empty bytestring")
        if key_len > 256:
            raise ValueError(f'key length should be less than 256, got {key_len}')

        # 使用 RC4-KSA 生成 S-box
        S = bytearray(range(256))
        j = 0
        for i in range(256):
            j = (j + S[i] + key[i % key_len]) % 256
            S[i], S[j] = S[j], S[i]

        # 使用 PRGA 生成密钥流
        meta_keystream = bytearray(256)
        for i, idx in enumerate(range(256), start=1):
            i %= 256
            si = S[i] % 256
            sj = S[(i + si) % 256] % 256
            K = S[(si + sj) % 256]
            meta_keystream[idx] = K

        self._meta_keystream = bytes(meta_keystream)

    def getkey(self, keyname: str = 'master') -> bytes:
        if keyname == 'master':
            return self._key

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike, /
                  ) -> Generator[int, None, None]:
        offset = toint(offset)
        nbytes = toint(nbytes)
        if offset < 0:
            raise ValueError("third argument 'offset' must be a non-negative integer")
        if nbytes < 0:
            raise ValueError("second argument 'nbytes' must be a non-negative integer")

        for i in range(offset, offset + nbytes):
            yield self._meta_keystream[i % 256]



# -*- coding: utf-8 -*-
from __future__ import annotations

import pickle
from pathlib import Path
from typing import Final

from ..miscutils import BINARIES_ROOTDIR

__all__ = ['KEY256_MAPPING']

MODULE_BINARIES_ROOTDIR = BINARIES_ROOTDIR / 'qmc' / Path(__file__).stem

# KEY256_MAPPING = [[]] * 256
#
# for i in range(128):
#     real_idx = (i * i + 27) % 256
#     if not KEY256_MAPPING[real_idx]:
#         KEY256_MAPPING[real_idx] = [i]
#     else:
#         KEY256_MAPPING[real_idx].append(i)
#
# KEY256_MAPPING 可使用以上代码生成
with open(MODULE_BINARIES_ROOTDIR / 'Key256MappingData', 'rb') as f:
    KEY256_MAPPING: Final[list[list[int]]] = pickle.load(f)



# -*- coding: utf-8 -*-
from __future__ import annotations

import array
import mmap
from os import PathLike
from typing import ByteString, Iterable, Iterator, Literal, Protocol, Sequence, SupportsBytes, SupportsIndex, SupportsInt, TypeVar, Union, runtime_checkable

__all__ = [
    'T',
    'KT',
    'VT',
    'PT',
    'RT',
    'AbleConvertToBytes',
    'AbleConvertToInt',
    'BytesLike',
    'IntegerLike',
    'WritableBuffer',
    'FilePath',
    'CipherProto',
    'StreamCipherProto',
    'KeyStreamBasedStreamCipherProto',
    'StreamCipherBasedCryptedIOProto'
]

T = TypeVar('T')
KT = TypeVar('KT')
VT = TypeVar('VT')
PT = TypeVar('PT')
RT = TypeVar('RT')

AbleConvertToBytes = Union[SupportsBytes, Iterable[int], Sequence[int]]
AbleConvertToInt = Union[SupportsInt, SupportsIndex]

BytesLike = Union[ByteString, AbleConvertToBytes]
IntegerLike = Union[int, AbleConvertToInt]
WritableBuffer = Union[bytearray, memoryview, array.array, mmap.mmap]

FilePath = Union[str, bytes, bytearray, PathLike]


@runtime_checkable
class CipherProto(Protocol):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        raise NotImplementedError

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        raise NotImplementedError

    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        raise NotImplementedError


@runtime_checkable
class StreamCipherProto(Protocol):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        raise NotImplementedError

    def encrypt(self, plaindata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        raise NotImplementedError

    def decrypt(self, cipherdata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        raise NotImplementedError


@runtime_checkable
class KeyStreamBasedStreamCipherProto(Protocol):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        raise NotImplementedError

    def keystream(self,
                  operation: Literal['encrypt', 'decrypt'],
                  nbytes: IntegerLike,
                  offset: IntegerLike = 0, /
                  ) -> Iterator[int]:
        raise NotImplementedError

    def encrypt(self, plaindata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        raise NotImplementedError

    def decrypt(self, cipherdata: BytesLike, offset: IntegerLike = 0, /) -> bytes:
        raise NotImplementedError


@runtime_checkable
class StreamCipherBasedCryptedIOProto(Protocol):
    @property
    def cipher(self) -> StreamCipherProto | KeyStreamBasedStreamCipherProto:
        raise NotImplementedError

    @property
    def master_key(self) -> bytes | None:
        raise NotImplementedError

    def read(self, size: IntegerLike = -1, /) -> bytes:
        raise NotImplementedError

    def write(self, data: BytesLike, /) -> int:
        raise NotImplementedError

    def seek(self, offset: IntegerLike, whence: IntegerLike = 0, /) -> int:
        raise NotImplementedError

    def tell(self) -> int:
        raise NotImplementedError

    def readable(self) -> bool:
        raise NotImplementedError

    def writable(self) -> bool:
        raise NotImplementedError

    def seekable(self) -> bool:
        raise NotImplementedError



# -*- coding: utf-8 -*-
from __future__ import annotations

from base64 import b64decode, b64encode
from math import tan
from typing import Iterable

from ..prototypes import CipherSkel
from ..stdciphers import TarsCppTCTEAWithModeCBC
from ..typedefs import BytesLike
from ..typeutils import tobytes

__all__ = [
    'make_core_key',
    'QMCv2KeyEncryptV1',
    'QMCv2KeyEncryptV2'
]


def make_core_key(salt: int, length: int) -> bytes:
    return bytes(int(abs(tan(salt + _ * 0.1) * 100)) for _ in range(length))


class QMCv2KeyEncryptV1(CipherSkel):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._core_key

    def __init__(self, key: BytesLike, /):
        self._core_key = tobytes(key)

        self._half_of_keysize = TarsCppTCTEAWithModeCBC.master_key_size // 2
        if len(self._core_key) != self._half_of_keysize:
            raise ValueError(f"invalid length of core key: "
                             f"should be {self._half_of_keysize}, not {len(self._core_key)}"
                             )

        self._key_buf = bytearray(TarsCppTCTEAWithModeCBC.master_key_size)
        for idx in range(TarsCppTCTEAWithModeCBC.blocksize):
            self._key_buf[idx << 1] = self._core_key[idx]

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        plaindata = tobytes(plaindata)
        if len(plaindata) < 8:  # 明文长度不可小于 8
            raise ValueError('invalid plaindata length: should be greater than or equal to 8, '
                             f'not {len(plaindata)}'
                             )

        recipe = plaindata[:8]
        payload = plaindata[8:]

        for idx in range(self._half_of_keysize):
            self._key_buf[(idx << 1) + 1] = recipe[idx]

        tea_cipher = TarsCppTCTEAWithModeCBC(self._key_buf, rounds=32)

        return recipe + tea_cipher.encrypt(payload)  # 返回值应当在 b64encode 后使用

    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        # cipherdata 应当为 b64decode 之后的结果
        cipherdata = tobytes(cipherdata)
        if len(cipherdata) < 8:  # 密文长度不可小于 8
            raise ValueError('invalid cipherdata length: should be greater than or equal to 8, '
                             f'not {len(cipherdata)}'
                             )

        recipe = cipherdata[:8]
        payload = cipherdata[8:]

        for idx in range(self._half_of_keysize):
            self._key_buf[(idx << 1) + 1] = recipe[idx]

        tea_cipher = TarsCppTCTEAWithModeCBC(self._key_buf, rounds=32)

        return recipe + tea_cipher.decrypt(payload, zero_check=True)


class QMCv2KeyEncryptV2(QMCv2KeyEncryptV1):
    def getkey(self, keyname: str = 'master') -> bytes | None:
        if keyname == 'master':
            return self._core_key
        elif keyname.startswith('garble'):
            garble_key_strno = keyname[6:]
            if garble_key_strno.isdecimal():
                garble_key_no = int(garble_key_strno)
                if 1 <= garble_key_no <= len(self._garble_ciphers):
                    return self._garble_ciphers[garble_key_no - 1].getkey()

    def __init__(self, key: BytesLike, garble_keys: Iterable[BytesLike], /) -> None:
        super().__init__(key)

        if not isinstance(garble_keys, Iterable):
            raise TypeError(
                f"argument 'garble_keys' must be an Iterable or bytes-like object, not {type(garble_keys).__name__}"
            )
        self._garble_ciphers: list[TarsCppTCTEAWithModeCBC] = []
        for idx, garble_key in enumerate(garble_keys, start=1):
            self._garble_ciphers.append(
                TarsCppTCTEAWithModeCBC(garble_key, rounds=32)
            )

    def encrypt(self, plaindata: BytesLike, /) -> bytes:
        # self.decrypt() 的反向操作
        result = tobytes(plaindata)

        result = b64encode(super().encrypt(result))

        for garble_cipher in self._garble_ciphers[::-1]:
            result = garble_cipher.encrypt(result)

        return result

    def decrypt(self, cipherdata: BytesLike, /) -> bytes:
        result = tobytes(cipherdata)
        # cipherdata 应当是 b64decode 之后，去除了开头 18 个字符的结果

        for garble_cipher in self._garble_ciphers:
            result = garble_cipher.decrypt(result, zero_check=True)
        # 结束循环时，cipherdata 应当是使用 QMCv2 Key Encrypt V1 加密的密钥
        return super().decrypt(b64decode(result, validate=True))



# -*- coding: utf-8 -*-
from __future__ import annotations

import warnings
from math import log10
from pathlib import Path
from typing import Callable, IO, NamedTuple, overload

from .kwmdataciphers import Mask32, Mask32FromRecipe
from ..exceptions import CrypterCreatingError, CrypterSavingError
from ..keyutils import make_salt
from ..miscutils import proberfuncfactory
from ..prototypes import EncryptedBytesIO
from ..typedefs import BytesLike, FilePath, IntegerLike, KeyStreamBasedStreamCipherProto, StreamCipherProto
from ..typeutils import isfilepath, tobytes, toint, verify_fileobj

warnings.filterwarnings(action='default', category=DeprecationWarning, module=__name__)

DIGIT_CHARS = b'0123456789'
ASCII_LETTER_CHARS = b'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'

__all__ = ['KWM', 'probe_kwm', 'KWMFileInfo']


class KWMFileInfo(NamedTuple):
    mask_recipe: bytes
    """组装主密钥所需的配方。"""
    cipher_data_offset: int
    """加密数据在文件中开始的位置。"""
    cipher_data_len: int
    """加密数据在文件中的长度。"""
    bitrate: int | None
    """加密数据（如果是音频）的比特率。"""
    suffix: str
    """加密数据的格式，可用于未加密形式文件的后缀。"""
    opener: Callable[[tuple[FilePath | IO[bytes], KWMFileInfo] | FilePath | IO[bytes], ...], KWM]
    """打开文件的方式，为一个可调对象，其会返回一个加密文件对象。"""
    opener_kwargs_required: tuple[str, ...]
    """通过 ``opener`` 打开文件时，所必需的关键字参数的名称。"""
    opener_kwargs_optional: tuple[str, ...]
    """通过 ``opener`` 打开文件时，可选的关键字参数的名称。

    此属性仅储存可能会影响 ``opener`` 行为的可选关键字参数；
    对 ``opener`` 行为没有影响的可选关键字参数不会出现在此属性中。
    """


@overload
def probe_kwm(filething: FilePath | IO[bytes], /) -> tuple[Path | IO[bytes], KWMFileInfo | None]:
    pass


@proberfuncfactory
def probe_kwm(filething, /):
    """探测源文件 ``filething`` 是否为一个 KWM 文件。

    返回一个 2 个元素长度的元组：第一个元素为 ``filething``；如果
    ``filething`` 是 KWM 文件，那么第二个元素为一个 ``KWMFileInfo`` 对象；否则为 ``None``。

    本方法的返回值可以用于 ``KWM.open()`` 的第一个位置参数。

    Args:
        filething: 源文件的路径或文件对象
    Returns:
        一个 2 个元素长度的元组：第一个元素为 filething；如果
        filething 是 KWM 文件，那么第二个元素为一个 KWMFileInfo 对象；否则为 None。
    """
    filething.seek(0, 0)

    header_data = filething.read(1024)
    cipher_data_offset = filething.tell()
    cipher_data_len = filething.seek(0, 2) - cipher_data_offset

    if not header_data.startswith(b'yeelion-kuwo'):
        return

    mask_recipe = header_data[24:32]

    bitrate = None
    bitrate_suffix_serialized = header_data[48:56].rstrip(b'\x00')
    bitrate_serialized_len = 0
    for byte in bitrate_suffix_serialized:
        if byte in DIGIT_CHARS:
            bitrate_serialized_len += 1
    if bitrate_serialized_len > 0:
        bitrate = int(bitrate_suffix_serialized[:bitrate_serialized_len]) * 1000

    suffix = None
    suffix_serialized = bitrate_suffix_serialized[bitrate_serialized_len:]
    suffix_serialized_len = 0
    for byte in suffix_serialized:
        if byte in ASCII_LETTER_CHARS:
            suffix_serialized_len += 1
    if suffix_serialized_len > 0:
        suffix = suffix_serialized[:suffix_serialized_len].decode('ascii')

    return KWMFileInfo(
        mask_recipe=mask_recipe,
        cipher_data_offset=cipher_data_offset,
        cipher_data_len=cipher_data_len,
        bitrate=bitrate,
        suffix=suffix,
        opener=KWM.open,
        opener_kwargs_required=('core_key',),
        opener_kwargs_optional=('master_key',)
    )


class KWM(EncryptedBytesIO):
    """基于 BytesIO 的 KWM 透明加密二进制流。

    所有读写相关方法都会经过透明加密层处理：
    读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

    调用读写相关方法时，附加参数 ``nocryptlayer=True``
    可绕过透明加密层，访问缓冲区内的原始加密数据。

    如果你要新建一个 KWM 对象，不要直接调用 ``__init__()``，而是使用构造器方法
    ``KWM.new()`` 和 ``KWM.open()`` 新建或打开已有 KWM 文件，
    使用已有 KWM 对象的 ``save()`` 方法将其保存到文件。
    """

    @property
    def acceptable_ciphers(self):
        return [Mask32FromRecipe]

    def __init__(self,
                 cipher: StreamCipherProto | KeyStreamBasedStreamCipherProto, /,
                 initial_bytes: BytesLike = b''
                 ) -> None:
        """基于 BytesIO 的 KWM 透明加密二进制流。

        所有读写相关方法都会经过透明加密层处理：
        读取时，返回解密后的数据；写入时，向缓冲区写入加密后的数据。

        调用读写相关方法时，附加参数 ``nocryptlayer=True``
        可绕过透明加密层，访问缓冲区内的原始加密数据。

        如果你要新建一个 KWM 对象，不要直接调用 ``__init__()``，而是使用构造器方法
        ``KWM.new()`` 和 ``KWM.open()`` 新建或打开已有 KWM 文件，
        使用已有 KWM 对象的 ``save()`` 方法将其保存到文件。

        Args:
            cipher: 要使用的 cipher，必须是一个 libtakiyasha.kwm.kwmdataciphers.Mask32/Mask32FromRecipe 对象
            initial_bytes: 内置缓冲区的初始数据
        """
        super().__init__(cipher, initial_bytes=initial_bytes)

        self._bitrate: int | None = None
        self._suffix: str | None = None

    @property
    def bitrate(self) -> int | None:
        """音频的比特率。如果要用作显示用途，需要除以 1000。

        不可设置为负数；如果不为 ``None``，其字面量长度与后缀 ``self.suffix`` 的长度不可超过 8。
        """
        return self._bitrate

    @bitrate.setter
    def bitrate(self, value: IntegerLike) -> None:
        """音频的比特率。如果要用作显示用途，需要除以 1000。

        不可设置为负数；如果不为 ``None``，其字面量长度与后缀 ``self.suffix`` 的长度不可超过 8。
        """
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'bitrate'. "
                f"Use `del self.bitrate` instead"
            )
        br = toint(value)
        if br < 0:
            raise ValueError(f"attribute 'bitrate' must be a non-netagive integer, not {value}")
        if self._suffix is None:
            max_bitrate_len = 8
        else:
            max_bitrate_len = 8 - len(self._suffix)
        bitrate_len = int(log10(br // 1000)) + 1
        if bitrate_len > max_bitrate_len:
            raise ValueError(f"attribute 'bitrate' must be less than {max_bitrate_len}, not {bitrate_len}")

        self._bitrate = br

    @bitrate.deleter
    def bitrate(self) -> None:
        """音频的比特率。本属性储存的是乘以 1000 后的结果。

        不可设置为负数；如果不为 ``None``，其整除 1000 后的字面量长度与后缀
        ``self.suffix`` 的长度之和不可大于 8。
        """
        self._bitrate = None

    @property
    def suffix(self) -> int | None:
        """加密数据对应的文件应当使用的后缀。由于不够精确，不建议使用。

        如果不为 None，其长度与比特率 ``self.bitrate`` 整除 1000 后的字面量长度之和不可大于 8。
        """
        return self._suffix

    @suffix.setter
    def suffix(self, value: str) -> None:
        """加密数据对应的文件应当使用的后缀。由于不够精确，不建议使用。

        如果不为 None，其长度与比特率 ``self.bitrate`` 整除 1000 后的字面量长度之和不可大于 8。
        """
        if value is None:
            raise TypeError(
                f"None cannot be assigned to attribute 'suffix'. "
                f"Use `del self.suffix` instead"
            )
        if not isinstance(value, str):
            raise TypeError(f"attribute 'suffix' must be str, not {type(value).__name__}")
        value = str(value)
        if self._bitrate is None:
            max_suffix_len = 8
        else:
            max_suffix_len = 8 - (int(log10(self._bitrate // 1000)) + 1)
        if len(value) > max_suffix_len:
            raise ValueError(
                f"attribute 'bitrate' must be less than {max_suffix_len}, not {len(value)}"
            )
        for char in (ord(_) for _ in value):
            if char not in ASCII_LETTER_CHARS:
                raise ValueError(
                    f"attribute 'suffix' can only contains digits and ascii letters, but '{chr(char)}' found"
                )
        self._suffix = value

    @suffix.deleter
    def suffix(self) -> None:
        """加密数据对应的文件应当使用的后缀。由于不够精确，不建议使用。

        如果不为 None，其长度与比特率 ``self.bitrate`` 整除 1000 后的字面量长度之和不可大于 8。
        """
        self._suffix = None

    @classmethod
    def from_file(cls,
                  kwm_filething: FilePath | IO[bytes], /,
                  core_key: BytesLike
                  ):
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``KWM.open()`` 代替。）

        打开一个 KWM 文件或文件对象 ``kwm_filething``。

        第一个位置参数 ``kwm_filething`` 可以是文件路径（``str``、``bytes``
        或任何拥有方法 ``__fspath__()`` 的对象）。``kwm_filething``
        也可以是一个文件对象，但必须可读、可跳转（``kwm_filething.seekable() == True``）。

        第二个参数 ``core_key`` 是必需的，它被用于还原和解密主密钥。
        """
        warnings.warn(
            DeprecationWarning(
                f'{cls.__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {cls.__name__}.open() instead.'
            )
        )

        return cls.open(kwm_filething, core_key=core_key)

    @classmethod
    def open(cls,
             filething_or_info: tuple[Path | IO[bytes], KWMFileInfo | None] | FilePath | IO[bytes], /,
             core_key: BytesLike = None,
             master_key: BytesLike = None
             ):
        """打开一个 KWM 文件，并返回一个 ``KWM`` 对象。

        第一个位置参数 ``filething_or_info`` 需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。

        ``filething_or_info`` 也可以接受 ``probe_kwm()`` 函数的返回值：
        一个包含两个元素的元组，第一个元素是源文件的路径或文件对象，第二个元素是源文件的信息。

        第二个参数 ``core_key`` 一般情况下是必需的，用于解密文件内嵌的主密钥。
        例外：如果你提供了第三个参数 ``mask``，那么它是可选的。

        第三个参数 ``mask`` 可选，如果提供，将会被作为主密钥使用，
        而文件内置的主密钥会被忽略，``core_key`` 也不再是必需参数。

        Args:
            filething_or_info: 源文件的路径或文件对象，或者 probe_kwm() 的返回值
            core_key: 核心密钥，用于生成文件内加密数据的主密钥
            master_key: 如果提供，将会被作为主密钥使用，而文件内置的主密钥会被忽略
        """
        if core_key is not None:
            core_key = tobytes(core_key)
        if master_key is not None:
            master_key = tobytes(master_key)

        def operation(fd: IO[bytes]) -> cls:
            fd.seek(1024, 0)
            initial_bytes = fd.read()

            if master_key is not None:
                cipher = Mask32(master_key)
            elif core_key is None:
                raise TypeError(
                    "argument 'core_key' is required to "
                    "generate the master key"
                )
            else:
                cipher = Mask32FromRecipe(fileinfo.mask_recipe, core_key)

            inst = cls(cipher, initial_bytes)
            inst._bitrate = fileinfo.bitrate
            inst._suffix = fileinfo.suffix

            return inst

        if isinstance(filething_or_info, tuple):
            filething_or_info: tuple[Path | IO[bytes], KWMFileInfo | None]
            if len(filething_or_info) != 2:
                raise TypeError(
                    "first argument 'filething_or_info' must be a file path, a file object, "
                    "or a tuple of probe_kwm() returns"
                )
            filething, fileinfo = filething_or_info
        else:
            filething, fileinfo = probe_kwm(filething_or_info)

        if fileinfo is None:
            raise CrypterCreatingError(
                f"{repr(filething)} is not a KWM file"
            )
        elif not isinstance(fileinfo, KWMFileInfo):
            raise TypeError(
                f"second element of the tuple must be KWMFileInfo or None, not {type(fileinfo).__name__}"
            )

        if isfilepath(filething):
            with open(filething, mode='rb') as fileobj:
                instance = operation(fileobj)
                instance._name = Path(filething)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_readable=True,
                                     verify_seekable=True
                                     )
            fileobj_sourcefile = getattr(fileobj, 'name', None)
            instance = operation(fileobj)

            if fileobj_sourcefile is not None:
                instance._name = Path(fileobj_sourcefile)

        return instance

    def to_file(self, kwm_filething: FilePath | IO[bytes] = None) -> None:
        """（已弃用，且将会在后续版本中删除。请尽快使用 ``KWM.save()`` 代替。）"""
        warnings.warn(
            DeprecationWarning(
                f'{type(self).__name__}.from_file() is deprecated, no longer used, '
                f'and may be removed in subsequent versions. '
                f'Use {type(self).__name__}.save() instead.'
            )
        )

        return self.save(kwm_filething)

    def save(self,
             filething: FilePath | IO[bytes] = None,
             newer_magic_header: bool = False
             ) -> None:
        """（实验性功能）将当前对象保存为一个新 KWM 文件。

        第一个参数 ``filething`` 是可选的，如果提供此参数，需要是一个文件路径或文件对象。
        可接受的文件路径类型包括：字符串、字节串、任何定义了 ``__fspath__()`` 方法的对象。
        如果是文件对象，那么必须可读且可寻址（其 ``seekable()`` 方法返回 ``True``）。
        如果未提供此参数，那么将会尝试使用当前对象的 ``source`` 属性；如果后者也不可用，则引发
        ``TypeError``。

        第二个参数 ``newer_magic_header`` 可选，如果为 ``True``，那么保存的文件会使用新版 KWM\
        文件的文件头 ``b'yeelion-kuwo'``；否则使用 ``b'yeelion-kuwo-tme'``。

        Args:
            filething: 目标文件的路径或文件对象
            newer_magic_header: 是否使用新版 KWM 文件使用的文件头
        """

        def operation(fd: IO[bytes]) -> None:
            recipe = self.cipher.getkey('original')
            if not recipe:
                raise CrypterSavingError('cannot store a non-standard recipe into a KWM file')

            fd.seek(0, 0)
            if newer_magic_header:
                fd.write(b'yeelion-kuwo')
            else:
                fd.write(b'yeelion-kuwo-tme')
            fd.seek(24, 0)
            fd.write(recipe)
            fd.seek(48, 0)
            if self._bitrate is not None:
                fd.write(str(self._bitrate // 1000).encode('ascii'))
            if self._suffix is not None:
                fd.write(self._suffix.encode('ascii'))
            fd.seek(1024, 0)
            fd.write(self.getvalue(nocryptlayer=True))

        if filething is None:
            if self.source is None:
                raise TypeError(
                    "attribute 'self.source' and argument 'filething' are empty, "
                    "don't know which file to save to"
                )
            filething = self.source

        if isfilepath(filething):
            with open(filething, mode='wb') as fileobj:
                return operation(fileobj)
        else:
            fileobj = verify_fileobj(filething, 'binary',
                                     verify_seekable=True,
                                     verify_writable=True
                                     )
            return operation(fileobj)

    @classmethod
    def new(cls, core_key: BytesLike):
        """返回一个空 KWM 对象。"""
        core_key = tobytes(core_key)

        recipe = make_salt(8)
        cipher = Mask32FromRecipe(recipe, core_key)

        return cls(cipher)



# -*- coding: utf-8 -*-
from __future__ import annotations

from threading import RLock
from typing import Any, Callable, IO, Literal, Type

from .typedefs import BytesLike, IntegerLike, PT, RT, T

__all__ = [
    'ClassInstanceProperty',
    'CachedClassInstanceProperty',
    'tobytes',
    'tobytearray',
    'toint',
    'isfilepath',
    'verify_fileobj'
]


class ClassInstanceProperty:
    """一个只读属性装饰器实现，通过此装饰器设置的只读属性，
    可同时在类及其实例中访问。

    通过此装饰器装饰的属性，在类未被实例化时，其
    ``setter()`` 和 ``deleter()`` 不可用。

    可用于设置一些需要在类及其实例中都保持只读的属性。
    """

    def __init__(self,
                 fget: Callable[[T], RT] = None,
                 fset: Callable[[T, PT], None] = None,
                 fdel: Callable[[T], None] = None,
                 doc: str = None
                 ) -> None:
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc
        self._attrname = ''

    def __set_name__(self, obj: T, attrname: str):
        self._attrname = attrname

    def __get__(self, obj: T, objtype: Type[T] = None) -> RT:
        if self.fget is None:
            raise AttributeError(f"unreadable attribute '{self._attrname}'")
        if obj is None:
            return self.fget(objtype)
        return self.fget(obj)

    def __set__(self, obj: T, value: PT) -> None:
        if self.fset is None:
            raise AttributeError(f"can't set attribute '{self._attrname}'")
        self.fset(obj, value)

    def __delete__(self, obj: T) -> None:
        if self.fdel is None:
            raise AttributeError(f"can't delete attribute '{self._attrname}'")
        self.fdel(obj)

    def getter(self, fget: Callable[[T], Any]):
        prop = type(self)(fget, self.fset, self.fdel, self.__doc__)
        prop._attrname = self._attrname
        return prop

    def setter(self, fset: Callable[[T, Any], Any]):
        prop = type(self)(self.fget, fset, self.fdel, self.__doc__)
        prop._attrname = self._attrname
        return prop

    def deleter(self, fdel: Callable[[T], Any]):
        prop = type(self)(self.fget, self.fset, fdel, self.__doc__)
        prop._attrname = self._attrname
        return prop


class CachedClassInstanceProperty(ClassInstanceProperty):
    """一个基于 ``ClassInstanceProperty`` 的只读属性装饰器实现，
    其功能和 ``ClassInstanceProperty`` 一致，但为 ``getter()`` 增加了缓存功能：

    首次访问属性后，本装饰器会记录上一次调用 ``getter()`` 返回的结果；

    如果在下一次访问此属性之前，没有通过 ``setter()`` 或 ``deleter()``
    修改属性，那么 ``getter()`` 会使用上次访问时的返回结果；

    如果通过 ``setter()`` 或 ``deleter()`` 修改了属性，那么 ``getter()`` 会
    再次调用 ``fget()``，记录并返回其返回值。
    """

    def __init__(self,
                 fget: Callable[[T], RT] = None,
                 fset: Callable[[T, PT], None] = None,
                 fdel: Callable[[T], None] = None,
                 doc: str = None
                 ) -> None:
        super().__init__(fget, fset, fdel, doc)
        self._lock = RLock()

    def __get__(self, obj: T, objtype: Type[T] = None) -> RT:
        if not hasattr(self, '_last_instance'):
            self._last_instance = obj
        elif self._last_instance != obj and hasattr(self, '_last_fget_returns'):
            del self._last_fget_returns

        if not hasattr(self, '_last_fget_returns'):
            with self._lock:
                if not hasattr(self, '_last_fget_returns'):
                    self._last_fget_returns = super().__get__(obj, objtype)
                    return self._last_fget_returns
        else:
            return self._last_fget_returns

    def __set__(self, obj: T, value: PT) -> None:
        if hasattr(self, '_last_fget_returns'):
            with self._lock:
                if hasattr(self, '_last_fget_returns'):
                    del self._last_fget_returns
        super().__set__(obj, value)

    def __delete__(self, obj: T) -> None:
        if hasattr(self, '_last_fget_returns'):
            with self._lock:
                if hasattr(self, '_last_fget_returns'):
                    del self._last_fget_returns
        super().__delete__(obj)


def tobytes(byteslike: BytesLike, /) -> bytes:
    """一个 ``bytes()`` 的包装器。

    本函数尝试将 ``byteslike`` 转换为 ``bytes``。

    和 ``bytes()`` 不一样，本函数不支持通过指定长度的方式创建 ``bytes``
    对象，即不接受整数类型 ``int`` 对象作为参数。如果输入此类对象，将会引发
    ``TypeError``。同时，本函数也不承担对字符串的编码。
    """
    if isinstance(byteslike, int) and not hasattr(byteslike, '__bytes__'):
        # 防止出现 bytes(114514) 这样的情况
        raise TypeError(f"a bytes-like object is required, not '{type(byteslike).__name__}'")
    else:
        return bytes(byteslike)


def tobytearray(byteslike: BytesLike, /) -> bytearray:
    """一个 ``bytearray()`` 的包装器。

    本函数尝试将 ``byteslike`` 转换为 ``bytearray``。

    和 ``bytearray()`` 不一样，本函数不支持通过指定长度的方式创建 ``bytearray``
    对象，即不接受整数类型 ``int`` 对象作为参数。如果输入此类对象，将会引发
    ``TypeError``。同时，本函数也不承担对字符串的编码。
    """
    if isinstance(byteslike, int) and not hasattr(byteslike, '__bytes__'):
        # 防止出现 bytearray(114514) 这样的情况
        raise TypeError(f"a bytes-like object is required, not '{type(byteslike).__name__}'")
    else:
        return bytearray(byteslike)


def toint(integerlike: IntegerLike, /) -> int:
    """一个 ``int()`` 的包装器。

    尽管 ``int()`` 可以转换浮点数，但是本函数不接受浮点数。如果输入此类对象，将会引发
    ``TypeError``。
    """
    if isinstance(integerlike, float):
        raise TypeError(f"'{type(integerlike).__name__}' object cannot be interpreted as an integer")
    else:
        return int(integerlike)


def isfilepath(obj, /) -> bool:
    """判断对象 ``obj`` 是否可以被视为文件路径。

    只有 ``str``、``bytes`` 类型，或者拥有 ``__fspath__``
    属性的对象，才会被视为文件路径。
    """
    return isinstance(obj, (str, bytes, bytearray)) or hasattr(obj, '__fspath__')


def verify_fileobj(fileobj: IO[str | bytes],
                   mode: Literal['text', 'binary'],
                   *,
                   verify_readable: bool = True,
                   verify_writable: bool = False,
                   verify_seekable: bool = True,
                   ) -> IO[str | bytes]:
    """

    Args:
        fileobj: 要验证的目标文件对象
        mode: fileobj 打开的模式，只能为 'text'（文本）和 'binary'（二进制）两个值
        verify_readable: 如果为真值，则验证 fileobj 是否可读
        verify_writable: 如果为真值，则验证 fileobj 是否可写
        verify_seekable: 如果为真值，则验证 fileobj 是否可任意改变读/写开始的位置
    """
    if mode == 'text':
        test_content = ''
        result_types = (str,)
    elif mode == 'binary':
        test_content = b'',
        result_types = (bytes, bytearray)
    elif isinstance(mode, str):
        raise ValueError("mode must be either 'text' or 'binary'")
    else:
        raise TypeError("verify_fileobj() argument 'mode' must be str, "
                        f"not {type(mode).__name__}"
                        )

    if bool(verify_readable):
        try:
            result = fileobj.read(0)
        except Exception as exc:
            if not hasattr(fileobj, 'read'):
                raise TypeError(f"{repr(fileobj)} is not a valid file object")
            raise ValueError(f"cannot read from file object {repr(fileobj)}") from exc
        else:
            if not isinstance(result, result_types):
                raise ValueError(f"file object '{repr(fileobj)}' is not open in {mode} mode")

    if verify_seekable:
        try:
            fileobj.seek(0, 1)
        except Exception as exc:
            if not hasattr(fileobj, 'seek'):
                raise TypeError(f"{repr(fileobj)} is not a valid file object")
            raise ValueError(f"cannot seek in file object {repr(fileobj)}") from exc

    if bool(verify_writable):
        try:
            fileobj.write(test_content)
        except TypeError as exc:
            raise ValueError(f"file object '{repr(fileobj)}' is not open in {mode} mode") from exc
        except Exception as exc:
            if not hasattr(fileobj, 'write'):
                raise TypeError(f"{repr(fileobj)} is not a valid file object")
            raise ValueError(f"cannot write to file object {repr(fileobj)}") from exc

    return fileobj
